//
//  USBClientLib.m
//  CppUserClient
//
//  Copyright (C) 2021 WCH.
//  Created by WCH TECH<tech@wch.cn> on 2021/9/17.
//

#import "CH37xLib.h"
#import <IOKit/IOKitLib.h>
#import <IOKit/IOReturn.h>


NSString * const DeviceWereConnectedNotification = @"DeviceWasConnectedNotification";
NSString * const DeviceWereDisconnectedNotification = @"DeviceWasDisconnectedNotification";

NSString * const ORSConnectedSerialPortsKey = @"ConnectedServiceKey";
NSString * const ORSDisconnectedSerialPortsKey = @"DisconnectedServiceKey";

@interface CH37xLib()

typedef NS_ENUM(NSUInteger, MessageType) {
    MessageType_AsyncWrite = 1,
    MessageType_SyncWrite = 2,
    MessageType_AsyncRead = 3,
    MessageType_SyncRead = 4,
    MessageType_ReadIOTOBuffer = 5,
    MessageType_GetDevice = 6,
    MessageType_SetTimeout = 7,
    MessageType_GetVersion = 8,
    MessageType_StartReadIO = 9,
    MessageType_StopReadIO = 10,
    MessageType_GetDeviceMsg = 11,
    MessageType_GetDeviceDesc = 12,
    MessageType_CloseClient = 13,
    MessageType_GetQueueSize = 14,
    MessageType_ResetWrite = 15,
    MessageType_SetBufUp = 16,
    MessageType_SetBufDown = 17,
    MessageType_GetLoactionID = 18,
};

@property (nonatomic, strong) NSMutableDictionary *endpoints;

@property (nonatomic, strong) NSMutableDictionary *isworkPoints;

@property (nonatomic, strong) NSMutableDictionary *queuePoints;

@property (nonatomic) io_iterator_t portPublishedNotificationIterator;
@property (nonatomic) io_iterator_t portTerminatedNotificationIterator;

@end

@implementation CH37xLib

mach_port_t machNotificationPort;
IONotificationPortRef fAsyncPort;
IONotificationPortRef terminationNotificationPort;
CFRunLoopRef globalRunLoop;
CFRunLoopSourceRef runLoopSource;

NSMutableArray *services;
NSMutableDictionary *connections;
NSMutableDictionary *locations;

static const char* dextIdentifier = "CH37xUSBDriver";

//io_connect_t connection = IO_OBJECT_NULL;

completionCallback readCallBack;
completionCallback writeCallBack;


- (instancetype)init
{
    self = [super init];
    if (self) {
        
        [self initData];
        
        globalRunLoop = CFRunLoopGetCurrent();
     
        fAsyncPort = IONotificationPortCreate(kIOMasterPortDefault);
        if (fAsyncPort == NULL)
        {
            printf("Failed to create notification port for application.\n");
        }
        
        machNotificationPort = IONotificationPortGetMachPort(fAsyncPort);
        if (machNotificationPort == 0)
        {
            printf("Failed to get mach notification port for application.\n");
        }
        
        runLoopSource = IONotificationPortGetRunLoopSource(fAsyncPort);
        if (runLoopSource == NULL)
        {
            printf("Failed to get run loop for application.\n");
        }
        
        // Establish our notifications in the run loop, so we can get callbacks.
        CFRunLoopAddSource(globalRunLoop, runLoopSource, kCFRunLoopDefaultMode);
        
    }
    
    return self;
}

- (void)initData {
    if (nil == services) {
        services = [NSMutableArray array];
    }
    
    if (nil == connections) {
        connections = [NSMutableDictionary dictionary];
    }
    
    if (nil == locations) {
        locations = [NSMutableDictionary dictionary];
    }
    
    if (nil == self.endpoints) {
        self.endpoints = [NSMutableDictionary dictionary];
    }
    
    if (nil == self.isworkPoints) {
        self.isworkPoints = [NSMutableDictionary dictionary];
    }
    
    if (nil == self.queuePoints) {
        self.queuePoints = [NSMutableDictionary dictionary];
    }
}

- (NSArray *)CH37xGetAllDevices {
    return services;
}

- (NSString *)CH37xGetLocationID:(NSString *)device {
    NSString *LocationID = [locations objectForKey:device];
    return LocationID;
}

- (NSInteger)CH37xScanDevices {
    
    kern_return_t ret = kIOReturnSuccess;
    io_iterator_t iterator = IO_OBJECT_NULL;
    io_service_t service = IO_OBJECT_NULL;
    io_connect_t connection = IO_OBJECT_NULL;
    
    ret = IOServiceAddMatchingNotification(fAsyncPort, kIOFirstMatchNotification, IOServiceNameMatching(dextIdentifier), DeviceAdded, (__bridge void *)(self), &iterator);
    if (ret != kIOReturnSuccess)
    {
        printf("Unable to find service for identifier with error: 0x%08x.\n", ret);
    }
    
    [services removeAllObjects];
    [connections removeAllObjects];
    
    self.portPublishedNotificationIterator = iterator;
    IOObjectRelease(iterator);

    while ((service = IOIteratorNext(self.portPublishedNotificationIterator)) != IO_OBJECT_NULL)
    {
        printf("发现设备：%d\n", service);
        
        ret = IOServiceOpen(service, mach_task_self_, 123, &connection);
        if (ret == kIOReturnSuccess) {
            NSString *serviceKey = [[NSString alloc] initWithFormat:@"%d", service];
            if (![services containsObject:serviceKey]) {
                [services addObject:serviceKey];
            }
 
            uint64_t LocationInfo[1] = {0};
            ret = CH37xGetLocationID(connection, LocationInfo, 1);
            if (ret == kIOReturnSuccess) {
                uint64_t locationID = LocationInfo[0];
                if (locationID > 0) {
                    
                    NSString *locationKey = [[NSString alloc] initWithFormat:@"%lld", locationID];
                    [locations setObject:locationKey forKey:serviceKey];
                }
            }
        
            [connections setObject:@(connection) forKey:serviceKey];
            printf("====>%d [0x%8llx]\n", connection, LocationInfo[0]);
        }
    }
    
    mach_port_t masterPort = IOMasterPort(MACH_PORT_NULL, &masterPort);
    CFMutableDictionaryRef matching_dictionary = IOServiceNameMatching(dextIdentifier);

    ret = IOServiceGetMatchingServices(masterPort, matching_dictionary, &iterator);
    if (ret != kIOReturnSuccess) {
        printf("Not Find the Matching service [0x%2x].\n", ret);
    }
    
    terminationNotificationPort = IONotificationPortCreate(kIOMasterPortDefault);
    CFRunLoopAddSource(globalRunLoop, IONotificationPortGetRunLoopSource(terminationNotificationPort), kCFRunLoopDefaultMode);

    
    ret = IOServiceAddMatchingNotification(terminationNotificationPort, kIOTerminatedNotification, IOServiceNameMatching(dextIdentifier), DeviceNotification, (__bridge void *)(self), &iterator);

    if (ret != kIOReturnSuccess)
    {
        printf("Unable to find service for identifier with error: 0x%08x.\n", ret);
    }

    self.portTerminatedNotificationIterator = iterator;
    IOObjectRelease(iterator);

    io_object_t device;
    while ((device = IOIteratorNext(self.portTerminatedNotificationIterator))) { IOObjectRelease(device); }
    
    return services.count;
}

- (io_connect_t)CH37xGetConnect:(NSInteger)index {
    if (index >= services.count) {
        return IO_OBJECT_NULL;
    }
    
    io_connect_t connect = IO_OBJECT_NULL;
    NSString *serviceKey = [[NSString alloc] initWithFormat:@"%@", [services objectAtIndex:index]];
    connect = (io_connect_t)[[connections objectForKey:serviceKey] intValue];
    if (connect == IO_OBJECT_NULL)
    {
        return IO_OBJECT_NULL;
    }
    return connect;
}

- (io_service_t)CH37xGetService:(NSInteger)index {
    if (index >= services.count) {
        return IO_OBJECT_NULL;
    }
    io_service_t service = IO_OBJECT_NULL;
    service = (io_service_t)[[services objectAtIndex:index] intValue];
    
    return service;
}


- (IOReturn)CH37xOpenDevice:(NSInteger)index{
    kern_return_t ret = kIOReturnSuccess;
    io_service_t service = IO_OBJECT_NULL;
    io_connect_t connection = IO_OBJECT_NULL;
    
    service = [self CH37xGetService:index];
    if (service == IO_OBJECT_NULL)
    {
        printf("Failed to match to device.\n");
        return EXIT_FAILURE;
    }
    
    connection = [self CH37xGetConnect:index];
    if (connection != IO_OBJECT_NULL) {
        printf("Device opened.\n");
        return kIOReturnSuccess;
    }
    
    // Open a connection to this user client as a server to that client, and store the instance in "service"
    ret = IOServiceOpen(service, mach_task_self_, 123, &connection);
    if (ret == kIOReturnSuccess) {
        printf("\tOpened service.\n");
        NSString *serviceKey = [[NSString alloc] initWithFormat:@"%d", service];
        [connections setObject:@(connection) forKey:serviceKey];
    }
    else {
        printf("\tFailed opening service with error: 0x%08x.\n", ret);
    }

    return ret;
}

- (IOReturn)CH37xCloseDevice:(NSInteger)index {
    
    if (index > services.count) {
        return EXIT_FAILURE;
    }
    
    IOReturn ret;
    io_service_t service = IO_OBJECT_NULL;
    io_connect_t connection = IO_OBJECT_NULL;
    
    service = [self CH37xGetService:index];
    if (service == IO_OBJECT_NULL)
    {
        printf("Failed to match to device.\n");
        return EXIT_FAILURE;
    }

    connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        printf("\tNO device opened.\n");
        return EXIT_FAILURE;
    }
    
    ret = IOConnectCallScalarMethod(connection, MessageType_CloseClient, NULL, 0, NULL, NULL);
    if (ret == kIOReturnSuccess) {
        NSString *serviceKey = [[NSString alloc] initWithFormat:@"%d", service];
        [connections removeObjectForKey:serviceKey];
        printf("\tClosed service.\n");
    }
    
    return ret;
}

#pragma -读操作

- (IOReturn)CH37xAsyncReadData:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress buffer:(void *)buffer size:(size_t)size callBack:(completionCallback)callBack {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    if (buffer == NULL) {
        return kIOReturnNoMemory;
    }
    
    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
    
    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Read EndPoint.\n");
        return kIOReturnError;
    }
    
    IOReturn ret;
    io_async_ref64_t asyncRef;

    if (!fAsyncPort) {
        return kIOReturnNotReady;
    }

    asyncRef[kIOAsyncCalloutFuncIndex] = (io_user_reference_t)AsyncCallback;
    asyncRef[kIOAsyncCalloutRefconIndex] = (io_user_reference_t)index;

    readCallBack = callBack;

    const uint32_t arraySize = 2;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;

    ret = IOConnectCallAsyncMethod(connection, MessageType_AsyncRead, machNotificationPort, asyncRef, kIOAsyncCalloutCount, input, arraySize, buffer, size, NULL, NULL, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("IOConnectCallAsyncMethod failed with error: 0x%2x.\n", ret);
    }
    
    return kIOReturnSuccess;
}

- (IOReturn)CH37xSyncReadData:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress buffer:(void *)buffer size:(size_t)size{
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    if (buffer == NULL) {
        return kIOReturnNoMemory;
    }
    
    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
    
    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Read EndPoint.\n");
        return kIOReturnError;
    }
    
    IOReturn ret = kIOReturnSuccess;
    const uint32_t arraySize = 2;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;
    
    ret = IOConnectCallMethod(connection, MessageType_SyncRead, input, arraySize, buffer, size, NULL, NULL, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Sync Read failed with error: 0x%08x.\n", ret);
    }
    
    return ret;
}


- (IOReturn)CH37xInterruptRead:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress buffer:(void *)buffer size:(size_t)size {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    if (buffer == NULL) {
        return kIOReturnNoMemory;
    }
    
    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
    
    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeInterrupt) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Read EndPoint.\n");
        return kIOReturnError;
    }
    
    IOReturn ret = kIOReturnSuccess;
    
    const uint32_t arraySize = 2;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;
    
    
    ret = IOConnectCallMethod(connection, MessageType_SyncRead, input, arraySize, buffer, size, NULL, NULL, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Interrupt Read failed with error: 0x%08x.\n", ret);
    }
    
    return ret;
}

- (IOReturn)CH37xStartBufUpload:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress  packageSize:(size_t)packageSize{

    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }

    IOReturn ret = kIOReturnSuccess;

    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];

    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Read EndPoint");
        return kIOReturnError;
    }

    const uint32_t arraySize = 3;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;
    input[2] = (uint64_t)packageSize;

    ret = IOConnectCallScalarMethod(connection, MessageType_StartReadIO, input, arraySize, NULL, NULL);

    if (ret != kIOReturnSuccess)
    {
        printf("Start Read failed with error: 0x%08x.\n", ret);
    }

    return ret;
}


- (IOReturn)CH37xStopBufUpload:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress  {

    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }

    IOReturn ret = kIOReturnSuccess;

    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];

    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Read EndPoint");
        return kIOReturnError;
    }

    const uint32_t arraySize = 2;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;

    ret = IOConnectCallScalarMethod(connection, MessageType_StopReadIO, input, arraySize, NULL, NULL);

    if (ret != kIOReturnSuccess)
    {
        printf("Stop Read failed with error: 0x%08x.\n", ret);
    }
    return ret;
}


- (IOReturn)CH37xReadEndP:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress buffer:(uint8_t *)buffer size:(size_t)size {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }

    IOReturn ret = kIOReturnSuccess;

    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];

    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Read EndPoint");
        return kIOReturnError;
    }

    const uint32_t inputSize = 2;
    uint64_t input[inputSize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;
    memset(buffer, 0, size);

    uint32_t outputSize = 2;
    uint64_t output[2] = {0};


    ret = IOConnectCallMethod(connection, MessageType_ReadIOTOBuffer, input, inputSize, buffer, size, output, &outputSize, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
//        printf("IOConnectCallStructMethod failed with error: 0x%08x.\n", ret);
    }

    return ret;
}

- (NSInteger)CH37xQueryBuf:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress {

    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    IOReturn ret = kIOReturnSuccess;

    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
    
    const uint32_t inputSize = 2;
    uint64_t input[inputSize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;

    uint32_t outputSize = 1;
    uint64_t output[1] = {0};
    
    uint64_t queueSize = 0;

    ret = IOConnectCallScalarMethod(connection, MessageType_GetQueueSize, input, inputSize, output, &outputSize);
    if (ret != kIOReturnSuccess)
    {
        printf("Get Buffer Count failed with error: 0x%08x.\n", ret);
    }
    
    queueSize = output[0];
    return queueSize;
}


#pragma -写操作

- (IOReturn)CH37xResetWrite:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    
    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
    
    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionOut << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Write EndPoint.\n");
        return kIOReturnError;
    }
    
    IOReturn ret;
    const uint32_t arraySize = 2;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;
    
    
    ret = IOConnectCallScalarMethod(connection, MessageType_ResetWrite, input, arraySize, NULL, NULL);
    if (ret != kIOReturnSuccess)
    {
        printf("Reset Write failed with error: 0x%08x.\n", ret);
    }

    return ret;
}

- (IOReturn)CH37xStartBufDownload:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress buffer:(void *)buffer size:(size_t)size needRead:(BOOL)isRead{
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    if (buffer == NULL) {
        return kIOReturnNoMemory;
    }
    
    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
    
    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionOut << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Write EndPoint.\n");
        return kIOReturnError;
    }
    
    IOReturn ret;
//    io_async_ref64_t asyncRef;
//
//    if (!fAsyncPort) {
//        return kIOReturnNotReady;
//    }
//
//    asyncRef[kIOAsyncCalloutFuncIndex] = (io_user_reference_t)AsyncCallback;
//    asyncRef[kIOAsyncCalloutRefconIndex] = (io_user_reference_t)index;
//    writeCallBack = callBack;
    
    const uint32_t arraySize = 3;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;
    input[2] = (uint64_t)isRead;
    
    ret = IOConnectCallMethod(connection, MessageType_AsyncWrite, input, arraySize, buffer, size, NULL, NULL, NULL, NULL);
    
//    ret = IOConnectCallAsyncMethod(connection, MessageType_AsyncWrite, machNotificationPort, &asyncRef[0], 2, input, arraySize, buffer, size, NULL, NULL, NULL, NULL);

    if (ret != kIOReturnSuccess)
    {
//        printf("Async Write failed with error: 0x%08x.\n", ret);
    }

    return ret;
}

- (IOReturn)CH37xSyncWriteData:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress buffer:(void *)buffer size:(size_t)size {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
    
    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionOut << kIOUSBEndpointDescriptorDirectionPhase))) {
        printf("Not Write EndPoint.\n");
        return kIOReturnError;
    }
    
    IOReturn ret = kIOReturnSuccess;
    
    const uint32_t arraySize = 2;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bEndpointAddress;
    input[1] = (uint64_t)bmAttributes;
    
    
    ret = IOConnectCallMethod(connection, MessageType_SyncWrite, input, arraySize, buffer, size, NULL, NULL, NULL, NULL);
    if (ret != kIOReturnSuccess)
    {
//        printf("Sync Write failed with error: 0x%08x.\n", ret);
    }
    
    return ret;
}

#pragma -其他操作

- (IOReturn)CH37xSetTimeout:(NSInteger)index readTimeout:(uint64_t)readTimeout writeTimeout:(uint64_t)writeTimeout {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    IOReturn ret = kIOReturnSuccess;
    
    const uint32_t arraySize = 2;
    const uint64_t input[arraySize] = {readTimeout, writeTimeout};
    
    ret = IOConnectCallScalarMethod(connection, MessageType_SetTimeout, input, arraySize, NULL, NULL);
    if (ret != kIOReturnSuccess)
    {
        printf("Set TimeOut failed with error: 0x%08x.\n", ret);
    }
    
    return ret;
}

- (IOReturn)CH37xGetVersion:(NSInteger)index buffer:(void *)buffer size:(size_t)size{

    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return kIOReturnInvalid;
    }
    
    if (buffer == NULL) {
        return kIOReturnNoMemory;
    }
    
    IOReturn ret = kIOReturnSuccess;
    
    ret = IOConnectCallStructMethod(connection, MessageType_GetVersion, buffer, size, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Get Version failed with error: 0x%08x.\n", ret);
    }

    return ret;
}

- (IOReturn)CH37xGetDeviceID:(NSInteger)index buffer:(uint64_t *)buffer size:(uint32_t)size{
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    if (buffer == NULL) {
        return kIOReturnNoMemory;
    }
    
    IOReturn ret = kIOReturnSuccess;
    
    ret = IOConnectCallScalarMethod(connection, MessageType_GetDevice, NULL, 0, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Get DeviceID failed with error: 0x%08x.\n", ret);
    }

    return ret;
}

- (NSArray *)CH37xGetDeviceMsg:(NSInteger)index{

//    uint8_t blength = (uint8_t)(result >> 8 * 6);
//    uint8_t bDescriptorType = (uint8_t)(result >> 8 * 5);
//    uint8_t bEndpointAddress = (uint8_t)(result >> 8 * 4);
//    uint8_t bmAttributes = (uint8_t)(result >> 8 * 3);
//    uint16_t wMaxPacketSize = (uint16_t)(result >> 8 * 1);
//    uint8_t bInterval = (uint8_t)(result >> 8 * 0);
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return nil;
    }
    
    IOReturn ret = kIOReturnSuccess;
    uint32_t size = 16;
    uint64_t *buffer = (uint64_t *)malloc(size * sizeof(uint64_t));
    memset(buffer, 0, size);
    
    NSMutableArray *endpoints = [[NSMutableArray alloc] init];
    
    ret = IOConnectCallScalarMethod(connection, MessageType_GetDeviceMsg, NULL, 0, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Get Device Message failed with error: 0x%08x.\n", ret);
    }
    
    for (int i = 0; i < size; i++) {
        
        uint64_t result = buffer[i];
        if (result == 0) {
            continue;
        }
        
        struct IOUSBEndpointDescriptor descriptor;
        descriptor.bLength = (uint8_t)(result >> 8 * 6);
        descriptor.bDescriptorType = (uint8_t)(result >> 8 * 5);
        descriptor.bEndpointAddress = (uint8_t)(result >> 8 * 4);
        descriptor.bmAttributes = (uint8_t)(result >> 8 * 3);
        descriptor.wMaxPacketSize = (uint16_t)(result >> 8 * 1);
        descriptor.bInterval = (uint8_t)(result >> 8 * 0);
        
        NSValue *value = [NSValue valueWithBytes:&descriptor objCType:@encode(struct IOUSBEndpointDescriptor)];
        [endpoints addObject:value];
        
        NSString *bEndpointAddress = [[NSString alloc] initWithFormat:@"%d", descriptor.bEndpointAddress];
        NSString *bmAttributes = [[NSString alloc] initWithFormat:@"%d", descriptor.bmAttributes];
        [self.endpoints setObject:bmAttributes forKey:bEndpointAddress];
    }

    return endpoints;
}

- (NSArray *)CH37xGetBulkInEndPoint:(NSInteger)index {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return nil;
    }
    
    IOReturn ret = kIOReturnSuccess;
    uint32_t size = 16;
    uint64_t *buffer = (uint64_t *)malloc(size * sizeof(uint64_t));
    memset(buffer, 0, size);
    
    NSMutableArray *endpoints = [[NSMutableArray alloc] init];
    
    ret = IOConnectCallScalarMethod(connection, MessageType_GetDeviceMsg, NULL, 0, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("IOConnectCallScalarMethod failed with error: 0x%08x.\n", ret);
    }
    
    for (int i = 0; i < size; i++) {
        
        uint64_t result = buffer[i];
        if (result == 0) {
            continue;
        }
        
        struct IOUSBEndpointDescriptor descriptor;
        descriptor.bLength = (uint8_t)(result >> 8 * 6);
        descriptor.bDescriptorType = (uint8_t)(result >> 8 * 5);
        descriptor.bEndpointAddress = (uint8_t)(result >> 8 * 4);
        descriptor.bmAttributes = (uint8_t)(result >> 8 * 3);
        descriptor.wMaxPacketSize = (uint16_t)(result >> 8 * 1);
        descriptor.bInterval = (uint8_t)(result >> 8 * 0);
        
        if (((descriptor.bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
            ((descriptor.bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase)) {
            NSValue *value = [NSValue valueWithBytes:&descriptor objCType:@encode(struct IOUSBEndpointDescriptor)];
            [endpoints addObject:value];
        }
    }

    return endpoints;
}

- (NSArray *)CH37xGetBulkOutEndPoint:(NSInteger)index {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return Nil;
    }
    
    IOReturn ret = kIOReturnSuccess;
    uint32_t size = 16;
    uint64_t *buffer = (uint64_t *)malloc(size * sizeof(uint64_t));
    memset(buffer, 0, size);
    
    NSMutableArray *endpoints = [[NSMutableArray alloc] init];
    
    ret = IOConnectCallScalarMethod(connection, MessageType_GetDeviceMsg, NULL, 0, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Get BulkOut failed with error: 0x%08x.\n", ret);
    }
    
    for (int i = 0; i < size; i++) {
        
        uint64_t result = buffer[i];
        if (result == 0) {
            continue;
        }
        
        struct IOUSBEndpointDescriptor descriptor;
        descriptor.bLength = (uint8_t)(result >> 8 * 6);
        descriptor.bDescriptorType = (uint8_t)(result >> 8 * 5);
        descriptor.bEndpointAddress = (uint8_t)(result >> 8 * 4);
        descriptor.bmAttributes = (uint8_t)(result >> 8 * 3);
        descriptor.wMaxPacketSize = (uint16_t)(result >> 8 * 1);
        descriptor.bInterval = (uint8_t)(result >> 8 * 0);
        
        
        if (((descriptor.bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
            ((descriptor.bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionOut << kIOUSBEndpointDescriptorDirectionPhase)) {
            NSValue *value = [NSValue valueWithBytes:&descriptor objCType:@encode(struct IOUSBEndpointDescriptor)];
            
            [endpoints addObject:value];
        }
    }

    return endpoints;
}

- (NSArray *)CH37xGetInterrputEndPoint:(NSInteger)index {
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return Nil;
    }
    
    IOReturn ret = kIOReturnSuccess;
    uint32_t size = 16;
    uint64_t *buffer = (uint64_t *)malloc(size * sizeof(uint64_t));
    memset(buffer, 0, size);
    
    NSMutableArray *endpoints = [[NSMutableArray alloc] init];
    
    ret = IOConnectCallScalarMethod(connection, MessageType_GetDeviceMsg, NULL, 0, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Get Intterrupt failed with error: 0x%08x.\n", ret);
    }
    
    for (int i = 0; i < size; i++) {
        
        uint64_t result = buffer[i];
        if (result == 0) {
            continue;
        }
        
        struct IOUSBEndpointDescriptor descriptor;
        descriptor.bLength = (uint8_t)(result >> 8 * 6);
        descriptor.bDescriptorType = (uint8_t)(result >> 8 * 5);
        descriptor.bEndpointAddress = (uint8_t)(result >> 8 * 4);
        descriptor.bmAttributes = (uint8_t)(result >> 8 * 3);
        descriptor.wMaxPacketSize = (uint16_t)(result >> 8 * 1);
        descriptor.bInterval = (uint8_t)(result >> 8 * 0);
        
        if (((descriptor.bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeInterrupt) &&
            ((descriptor.bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase)) {
            NSValue *value = [NSValue valueWithBytes:&descriptor objCType:@encode(struct IOUSBEndpointDescriptor)];
            [endpoints addObject:value];
        }
    }

    return endpoints;
}

- (IOReturn)CH37xGetDeviceDesc:(NSInteger)index buffer:(uint64_t *)buffer {
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    if (buffer == NULL) {
        return kIOReturnNoMemory;
    }
    
    uint32_t size = 12;
    
    IOReturn ret = kIOReturnSuccess;
    
    ret = IOConnectCallScalarMethod(connection, MessageType_GetDeviceDesc, NULL, 0, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Get Deivce Desc failed with error: 0x%08x.\n", ret);
    }

    
    return ret;
}

- (IOReturn)CH37xSetBufUpload:(NSInteger)index bufferSize:(size_t)bufferSize {
    
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    IOReturn ret;
    const uint32_t arraySize = 1;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bufferSize;
    
    ret = IOConnectCallScalarMethod(connection, MessageType_SetBufUp, input, arraySize, NULL, NULL);
    if (ret != kIOReturnSuccess)
    {
        printf("Reset Write failed with error: 0x%08x.\n", ret);
    }

    return ret;
}

- (IOReturn)CH37xSetBufDownload:(NSInteger)index bufferSize:(size_t)bufferSize {
    io_connect_t connection = [self CH37xGetConnect:index];
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    IOReturn ret;
    const uint32_t arraySize = 1;
    uint64_t input[arraySize] = {0};
    input[0] = (uint64_t)bufferSize;
    
    ret = IOConnectCallScalarMethod(connection, MessageType_SetBufDown, input, arraySize, NULL, NULL);
    if (ret != kIOReturnSuccess)
    {
        printf("Reset Write failed with error: 0x%08x.\n", ret);
    }

    return ret;
}


void AsyncCallback(void* refcon, IOReturn result, void** args, uint32_t numArgs)
{
    const char* funcName = NULL;
    uint64_t* arrArgs = (uint64_t*)args;
    uint64_t index = (uint64_t)refcon;
    
    switch (arrArgs[0])
    {
        case 1:
        {
            funcName = "'Async Write'";
            if (writeCallBack) {
                writeCallBack(index);
            }
            
        } break;

        case 2:
        {
            funcName = "'Async Read'";
            if (readCallBack) {
                readCallBack(index);
            }
            
        } break;

        default:
        {
            funcName = "UNKNOWN";
        } break;
    }
}

void DeviceNotification(void *refCon, io_iterator_t iterator) {
    printf("remove device.\n");
    CH37xLib *manager = (__bridge CH37xLib *)refCon;
    if (![manager isKindOfClass:[CH37xLib class]])
    {
        return;
    }
    [manager CH37xWereTerminated:iterator];
}

void DeviceAdded(void *refCon, io_iterator_t iterator) {
    printf("add device.\n");
    CH37xLib *manager = (__bridge CH37xLib *)refCon;
    if (![manager isKindOfClass:[CH37xLib class]])
    {
        return;
    }
    [manager CH37xsWerePublished:iterator];
}

- (void)CH37xsWerePublished:(io_iterator_t)iterator;
{
    kern_return_t ret = kIOReturnSuccess;
    io_service_t service = IO_OBJECT_NULL;
    io_connect_t connection = IO_OBJECT_NULL;
   
    while ((service = IOIteratorNext(iterator)) != IO_OBJECT_NULL)
    {
        printf("find service：%d\n", service);
        
        ret = IOServiceOpen(service, mach_task_self_, 123, &connection);
        if (ret == kIOReturnSuccess) {
            
            NSString *serviceKey = [[NSString alloc] initWithFormat:@"%d", service];
            
            if (![services containsObject:serviceKey]) {
                [services addObject:serviceKey];
            }
        
            uint64_t LocationInfo[1] = {0};
            ret = CH37xGetLocationID(connection, LocationInfo, 1);
            if (ret == kIOReturnSuccess) {
                uint64_t locationID = LocationInfo[0];
                if (locationID > 0) {
                    NSString *locationKey = [[NSString alloc] initWithFormat:@"%lld", locationID];
                    [locations setObject:locationKey forKey:serviceKey];
                    
                    if (self.delegate && [self.delegate respondsToSelector:@selector(CH37xDeviceConnect:)]) {
                        [self.delegate CH37xDeviceConnect:locationKey];
                    }
                    
                }
            }
            [connections setObject:@(connection) forKey:serviceKey];
            printf("====>%d [0x%8llx]\n", connection, LocationInfo[0]);
        }
    }
}

- (void)CH37xWereTerminated:(io_iterator_t)iterator;
{
    io_service_t service = IO_OBJECT_NULL;
   
    while ((service = IOIteratorNext(iterator)) != IO_OBJECT_NULL)
    {
        printf("find service：%d\n", service);
        NSString *serviceKey = [[NSString alloc] initWithFormat:@"%d", service];
        [services removeObject:serviceKey];
        
        
        NSString *locationKey = [locations objectForKey:serviceKey];
        [connections removeObjectForKey:serviceKey];
        [locations removeObjectForKey:serviceKey];
        
        if (self.delegate && [self.delegate respondsToSelector:@selector(CH37xDeviceDisConnect:)]) {
            [self.delegate CH37xDeviceDisConnect:locationKey];
        }
     
    }
}

IOReturn CH37xGetLocationID(io_connect_t connection, uint64_t *buffer, uint32_t size) {
    if (connection == IO_OBJECT_NULL) {
        return EXIT_FAILURE;
    }
    
    IOReturn ret = kIOReturnSuccess;
    
    ret = IOConnectCallScalarMethod(connection, MessageType_GetLoactionID, NULL, 0, buffer, &size);
    if (ret != kIOReturnSuccess)
    {
        printf("Get LocationID failed with error: 0x%08x.\n", ret);
    }
    
    return ret;
}


- (void)setPortPublishedNotificationIterator:(io_iterator_t)iterator
{
    if (iterator != _portPublishedNotificationIterator)
    {
        if (_portPublishedNotificationIterator) IOObjectRelease(_portPublishedNotificationIterator);
        
        _portPublishedNotificationIterator = iterator;
        IOObjectRetain(_portPublishedNotificationIterator);
    }
}

- (void)setPortTerminatedNotificationIterator:(io_iterator_t)iterator
{
    if (iterator != _portTerminatedNotificationIterator)
    {
        if (_portTerminatedNotificationIterator) IOObjectRelease(_portTerminatedNotificationIterator);
        
        _portTerminatedNotificationIterator = iterator;
        IOObjectRetain(_portTerminatedNotificationIterator);
    }
}

@end

//    共享内存
//    mach_vm_address_t vmaddress = (uint64_t)buffer;
//    mach_vm_size_t    vmsize = (uint64_t)size;

//    ret = IOConnectMapMemory(connection, 2, mach_task_self(), &vmaddress, &vmsize, kIOMapAnywhere);
//    if (ret != kIOReturnSuccess) {
//        printf("IOConnectMapMemory failed with error: 0x%08x.\n", ret);
//    }


//- (IOReturn)CH37xStartReadIO:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress  {
//
//    NSString *key = [[NSString alloc] initWithFormat:@"%ld", (long)index];
//    NSNumber *number = [connections objectForKey:key];
//    io_connect_t connection = [number intValue];
//    if (connection == IO_OBJECT_NULL) {
//        return EXIT_FAILURE;
//    }
//
//    IOReturn ret = kIOReturnSuccess;
//
//    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
//    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
//
//    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
//         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
//        printf("Not Read EndPoint");
//        return kIOReturnError;
//    }
//
//    const uint32_t arraySize = 2;
//    uint64_t input[arraySize] = {0};
//    input[0] = (uint64_t)bEndpointAddress;
//    input[1] = (uint64_t)bmAttributes;
//
//    ret = IOConnectCallScalarMethod(connection, MessageType_StartReadIO, input, arraySize, NULL, NULL);
//
//    if (ret != kIOReturnSuccess)
//    {
//        printf("IOConnectCallScalarMethod failed with error: 0x%08x.\n", ret);
//    }
//
//    return ret;
//}
//
//
//- (IOReturn)CH37xStopReadIO:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress  {
//
//    NSString *key = [[NSString alloc] initWithFormat:@"%ld", (long)index];
//    NSNumber *number = [connections objectForKey:key];
//    io_connect_t connection = [number intValue];
//    if (connection == IO_OBJECT_NULL) {
//        return EXIT_FAILURE;
//    }
//
//    IOReturn ret = kIOReturnSuccess;
//
//    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
//    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
//
//    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
//         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
//        printf("Not Read EndPoint");
//        return kIOReturnError;
//    }
//
//    const uint32_t arraySize = 2;
//    uint64_t input[arraySize] = {0};
//    input[0] = (uint64_t)bEndpointAddress;
//    input[1] = (uint64_t)bmAttributes;
//
//    ret = IOConnectCallScalarMethod(connection, MessageType_StopReadIO, input, arraySize, NULL, NULL);
//
//    if (ret != kIOReturnSuccess)
//    {
//        printf("IOConnectCallScalarMethod failed with error: 0x%08x.\n", ret);
//    }
//
//    return ret;
//}
//
//- (size_t)CH37xReadIOTOBuffer:(NSInteger)index bEndpointAddress:(uint8_t)bEndpointAddress buffer:(void *)buffer size:(size_t)size {
//
//    NSString *key = [[NSString alloc] initWithFormat:@"%ld", (long)index];
//    NSNumber *number = [connections objectForKey:key];
//    io_connect_t connection = [number intValue];
//    if (connection == IO_OBJECT_NULL) {
//        return -1;
//    }
//
//    IOReturn ret = kIOReturnSuccess;
//
//    NSString *addressStr = [[NSString alloc] initWithFormat:@"%d", bEndpointAddress];
//    uint8_t bmAttributes = [[self.endpoints objectForKey:addressStr] intValue];
//
//    if (!(((bmAttributes & kIOUSBEndpointDescriptorTransferType) == kIOUSBEndpointTypeBulk) &&
//         ((bEndpointAddress & kIOUSBEndpointDescriptorDirection) == kIOUSBEndpointDirectionIn << kIOUSBEndpointDescriptorDirectionPhase))) {
//        printf("Not Read EndPoint");
//        return -1;
//    }
//
//    const uint32_t inputSize = 2;
//    uint64_t input[inputSize] = {0};
//    input[0] = (uint64_t)bEndpointAddress;
//    input[1] = (uint64_t)bmAttributes;
//    memset(buffer, 0, size);
//
//    uint32_t outputSize = 2;
//    uint64_t output[2] = {0};
//
//
//    ret = IOConnectCallMethod(connection, MessageType_ReadIOTOBuffer, input, inputSize, buffer, size, output, &outputSize, buffer, &size);
//    if (ret != kIOReturnSuccess)
//    {
//        printf("IOConnectCallStructMethod failed with error: 0x%08x.\n", ret);
//
//        return -1;
//    }
//
//    return output[0];
//}
